#######################################################################
#                            Flux methods                             #
#######################################################################



@with_signature function flux(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho = q
  flx_q   = q^2 / rho + p
  flx_E   = q / rho * (E + p)
  @returns flx_rho, flx_q, flx_E
end



@with_signature function speed(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function internal_energy(::EulerEquation)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, eps
  @unpack eos = equation
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function entropy(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, q, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation)
  @accepts rho, q, S
  @unpack eos  = equation
  flx_S = q / rho * S
  @returns flx_S
end



#######################################################################
#                                 2D                                  #
#######################################################################


LLF(nfl,nfr,ul,ur,v) = 1/2 * (nfl + nfr - v * (ur - ul))


@with_signature function internal_energy(::EulerEquation2d)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, E, p
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho_x = qx
  flx_rho_y = qy
  flx_qx_x = qx^2/rho + p
  flx_qx_y = qx*qy / rho
  flx_qy_y = qy^2/rho + p
  flx_E_x  = qx / rho * (E + p)
  flx_E_y  = qy / rho * (E + p)
  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y, flx_qy_y, flx_E_x, flx_E_y
end


@with_signature function llf_rho(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, v
  @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  vmax = max(lhs_v, rhs_v)
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)
  @returns nflx_rho
end
@with_signature function llf_qx(equation::EulerEquation2d)
  @accepts Prefix(lhs), qx, flx_qx_x, flx_qx_y, v
  @accepts Prefix(rhs), qx, flx_qx_x, flx_qx_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  vmax = max(lhs_v, rhs_v)
  nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)
  @returns nflx_qx
end
@with_signature function llf_qy(equation::EulerEquation2d)
  @accepts Prefix(lhs), qy, flx_qy_x, flx_qy_y, v
  @accepts Prefix(rhs), qy, flx_qy_x, flx_qy_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  vmax = max(lhs_v, rhs_v)
  nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)
  @returns nflx_qy
end
@with_signature function llf_E(equation::EulerEquation2d)
  @accepts Prefix(lhs), E, flx_E_x, flx_E_y, v
  @accepts Prefix(rhs), E, flx_E_x, flx_E_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  vmax = max(lhs_v, rhs_v)
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)
  @returns nflx_E
end


@with_signature function llf(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_x, flx_qy_y, E, flx_E_x, flx_E_y, v
  @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, qx, flx_qx_x, flx_qx_y, qy, flx_qy_x, flx_qy_y, E, flx_E_x, flx_E_y, v
  @accepts nx, ny

  vmax = max(lhs_v, rhs_v)

  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)

  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)

  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end


@with_signature function maxspeed(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function entropy(equation::EulerEquation2d{EquationOfState.IdealGas})
  @accepts rho, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, S
  @unpack eos  = equation
  flx_S_x = qx / rho * S
  flx_S_y = qy / rho * S
  @returns flx_S_x, flx_S_y
end

#######################################################################
#                         LDG implementation                          #
#######################################################################


@with_signature function ldg_flux(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), rho, q, E
  flx_ldg_rho, flx_ldg_q, flx_ldg_E = rho, q, E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function ldg_speed(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), rho, q, E
  v_ldg_rho, v_ldg_q, v_ldg_E = 1, 1, 1
  @returns v_ldg_rho, v_ldg_q, v_ldg_E
end


@with_signature function av_flux(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), smoothed_mu
  flx_ldg_rho, flx_ldg_q, flx_ldg_E =
    smoothed_mu * ldg_rho, smoothed_mu * ldg_q, smoothed_mu * ldg_E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function av_speed(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), smoothed_mu
  v_av_rho, v_av_q, v_av_E = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_av_rho, v_av_q, v_av_E
end


@with_signature function ldg_nflux(eq::EulerEquation2d)
  @accepts Prefix(lhs), rho, qx, qy, E
  @accepts Prefix(rhs), rho, qx, qy, E
  @accepts nx, ny

  nflx_rho_x = 0.5 * nx * (lhs_rho + rhs_rho)
  nflx_rho_y = 0.5 * ny * (lhs_rho + rhs_rho)

  nflx_qx_x = 0.5 * nx * (lhs_qx + rhs_qx)
  nflx_qx_y = 0.5 * ny * (lhs_qx + rhs_qx)

  nflx_qy_x = 0.5 * nx * (lhs_qy + rhs_qy)
  nflx_qy_y = 0.5 * ny * (lhs_qy + rhs_qy)

  nflx_E_x = 0.5 * nx * (lhs_E + rhs_E)
  nflx_E_y = 0.5 * ny * (lhs_E + rhs_E)

  @returns nflx_rho_x, nflx_rho_y, nflx_qx_x, nflx_qx_y,
           nflx_qy_x, nflx_qy_y, nflx_E_x, nflx_E_y
end


@with_signature function av_flux_2d(eq::EulerEquation2d)
  @accepts flx_rho_x, flx_rho_y, ldg_rho_x, ldg_rho_y,
           flx_qx_x, flx_qx_y, ldg_qx_x, ldg_qx_y,
           flx_qy_x, flx_qy_y, ldg_qy_x, ldg_qy_y,
           flx_E_x, flx_E_y, ldg_E_x, ldg_E_y,
           smoothed_mu

  flx_rho_x += smoothed_mu * ldg_rho_x
  flx_rho_y += smoothed_mu * ldg_rho_y

  flx_qx_x += smoothed_mu * ldg_qx_x
  flx_qx_y += smoothed_mu * ldg_qx_y

  flx_qy_x += smoothed_mu * ldg_qy_x
  flx_qy_y += smoothed_mu * ldg_qy_y

  flx_E_x += smoothed_mu * ldg_E_x
  flx_E_y += smoothed_mu * ldg_E_y

  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y,
           flx_qy_x, flx_qy_y, flx_E_x, flx_E_y
end
@with_signature function av_nflux_2d(eq::EulerEquation2d)
  @accepts Prefix(lhs), ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y, ldg_E_x, ldg_E_y, smoothed_mu
  @accepts Prefix(rhs), ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y, ldg_E_x, ldg_E_y, smoothed_mu
  @accepts nflx_rho, nflx_qx, nflx_qy, nflx_E, nx, ny

  lhs_nflx = nx*lhs_ldg_rho_x + ny*lhs_ldg_rho_y
  rhs_nflx = nx*rhs_ldg_rho_x + ny*rhs_ldg_rho_y
  nflx_rho += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_qx_x + ny*lhs_ldg_qx_y
  rhs_nflx = nx*rhs_ldg_qx_x + ny*rhs_ldg_qx_y
  nflx_qx += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_qy_x + ny*lhs_ldg_qy_y
  rhs_nflx = nx*rhs_ldg_qy_x + ny*rhs_ldg_qy_y
  nflx_qy += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  lhs_nflx = nx*lhs_ldg_E_x + ny*lhs_ldg_E_y
  rhs_nflx = nx*rhs_ldg_E_x + ny*rhs_ldg_E_y
  nflx_E += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end
