"""
    compute_rhs_weak_form!(rhs, f, nf_lhs, nf_rhs, mesh)
    compute_rhs_weak_form!(rhs, f, s, nf_lhs, nf_rhs, mesh)
"""
function compute_rhs_weak_form!(rhs, f, nf, mesh::Mesh1d)
  @unpack invjac, element = mesh
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  @unpack K = mesh
  shape        = layout(mesh)
  mat_rhs      = vreshape(rhs, shape)
  mat_f        = vreshape(f, shape)
  mul!(mat_rhs, MDM, mat_f)
  nf_lhs       = view(nf, 1:K)
  nf_rhs       = view(nf, K+1:2*K)
  @. mat_rhs  -= (nf_rhs' * Ml_rhs + #= minus contained in normal vector =# nf_lhs' * Ml_lhs)
  mat_rhs    .*= invjac
  return
end
function compute_rhs_weak_form!(rhs, f, s, nf, mesh::Mesh1d)
  compute_rhs_weak_form!(rhs, f, nf, mesh)
  rhs .+= s
  return
end

function compute_rhs_weak_form!(rhs, fx, fy, nf, mesh::Mesh2d)
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  @inbounds for k = 1:n_cells(mesh.tree)

    # bulk
    idxs_bulk = cellindices(mesh, k)
    v_rhs     = view(rhs,  idxs_bulk)
    v_fx      = view(fx,   idxs_bulk)
    v_fy      = view(fy,   idxs_bulk)
    v_dxdX    = view(dxdX, idxs_bulk)
    v_dydY    = view(dydY, idxs_bulk)

    # faces
    idxs_lhs, idxs_rhs, idxs_down, idxs_up = faceindices(mesh, k)

    v_nf_lhs  = view(nf, idxs_lhs)
    v_nf_rhs  = view(nf, idxs_rhs)
    v_nf_down = view(nf, idxs_down)
    v_nf_up   = view(nf, idxs_up)

    # TODO Benchmark and see whether LoopVectorization can used here.
    @inbounds for j=1:Ny, i=1:Nx
      rhsij = 0
      dxdXij = v_dxdX[i,j]
      dydYij = v_dydY[i,j]
      for k = 1:Nx
        rhsij += MDM[i,k] * v_fx[k,j]
      end
      v_rhs[i,j] = rhsij * dxdXij
      rhsij = 0
      for k = 1:Ny
        rhsij += MDM[j,k] * v_fy[i,k]
      end
      v_rhs[i,j] += rhsij * dydYij
      v_rhs[i,j] -= (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i])  * dxdXij +
                    (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
    end

  end
  return
end
function compute_rhs_weak_form!(rhs, fx, fy::Real, nf, mesh::Mesh2d)
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  @inbounds for k = 1:n_cells(mesh.tree)

    # bulk
    idxs_bulk = cellindices(mesh, k)
    v_rhs     = view(rhs,  idxs_bulk)
    v_fx      = view(fx,   idxs_bulk)
    v_dxdX    = view(dxdX, idxs_bulk)

    # faces
    idxs_lhs, idxs_rhs, _, _ = faceindices(mesh, k)

    v_nf_lhs  = view(nf, idxs_lhs)
    v_nf_rhs  = view(nf, idxs_rhs)

    # TODO Benchmark and see whether LoopVectorization can used here.
    @inbounds for j=1:Ny, i=1:Nx
      rhsij = 0
      dxdXij = v_dxdX[i,j]
      for k = 1:Nx
        rhsij += MDM[i,k] * v_fx[k,j]
      end
      v_rhs[i,j] = rhsij * dxdXij
      v_rhs[i,j] -= (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i])  * dxdXij
    end

  end
  return
end
function compute_rhs_weak_form!(rhs, fx::Real, fy, nf, mesh::Mesh2d)
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  @inbounds for k = 1:n_cells(mesh.tree)

    # bulk
    idxs_bulk = cellindices(mesh, k)
    v_rhs     = view(rhs,  idxs_bulk)
    v_fy      = view(fy,   idxs_bulk)
    v_dydY    = view(dydY, idxs_bulk)

    # faces
    _, _, idxs_down, idxs_up = faceindices(mesh, k)

    v_nf_down = view(nf, idxs_down)
    v_nf_up   = view(nf, idxs_up)

    # TODO Benchmark and see whether LoopVectorization can used here.
    @inbounds for j=1:Ny, i=1:Nx
      rhsij = 0
      dydYij = v_dydY[i,j]
      for k = 1:Ny
        rhsij += MDM[j,k] * v_fy[i,k]
      end
      v_rhs[i,j] += rhsij * dydYij
      v_rhs[i,j] -= (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
    end

  end
  return
end
function compute_rhs_weak_form!(rhs, fx, fy, s, nf, mesh::Mesh2d)
  compute_rhs_weak_form!(rhs, fx, fy, nf, mesh)
  rhs .+= s
  return
end
