"""
    @with_signature

Macro to extract `accepted` and `returned` arguments from a function and make them
programmatically accessible. Functions defined using this macro can be used with
the `broadcast_volume!, broadcast_faces!, broadcast_bdrys` interfaces.


# Example
```
@with_signature function fn(eq::Equation)
  @accepts State(u1, u2), a, b
  # compute values rhs_u1, rhs_u2
  @returns rhs_u1, rhs_u2
end
```
translates into
```
function fn(args::NTuple{4,Float64}, eq::Equation)
  (u1, u2, a, b) = args
  # compute values rhs_u1, rhs_u2
  (rhs_u1, rhs_u2)
end
accepts(::typeof(fn), ::Equation) = (:u1, :u2, :a, :b)
returns(::typeof(fn), ::Equation) = (:rhs_u1, :rhs_u2)
state_indices(::typeof(fn), ::Equation) = (1,2)
has_signature(::typeof(fn), ::Equation) = true
```
"""
macro with_signature(ex)
  return with_signature(ex)
end


"""
See @with_signature
"""
macro accepts(ex); ex; end
macro returns(ex); ex; end


_is_state(ex) = ex isa Expr && ex.head === :call && ex.args[1] === :State
_is_tuple(ex) = ex isa Expr && ex.head === :tuple


# gather list of all variables and a list of indices of variables
# tagged with State and unpack optional Prefix names, e.g.
# accepts(:(State(a, b), c, d, State(e), f))
# ->
# (:a, :b, :c, :d, :e, :f), (1, 2, 5), nothing
# oor
# accepts(:(Prefix(lhs), c, d, State(e), f))
# ->
# (:c, :d, :e, :f), (), :lhs
function accepts(ex)
  allsyms, i_states, prefix = if ex isa Symbol
    (ex,), (), nothing
  elseif _is_state(ex)
    vars = ex.args[2:end]
    vars, collect(1:length(vars)), nothing
  elseif _is_tuple(ex)
    vars = ex.args[1:end]
    allsyms, i_states = Symbol[], Int64[]
    prefix = if vars[1] isa Expr && vars[1].head === :call && vars[1].args[1] === :Prefix
      if length(vars[1].args) != 2
        error("@accepts: invalid use of 'Prefix' in '$(vars[1])', use as '@accepts Prefix(lhs), a, b'")
      end
      p = vars[1].args[2]
      if !(p isa Symbol)
        error("@accepts: expected a symbol as prefix, found '$(vars[1])'")
      end
      popfirst!(vars)
      p
    else
      nothing
    end
    last_i = 1
    for v in vars
      if _is_state(v)
        vs = v.args[2:end]
        push!(allsyms, vs...)
        append!(i_states, collect(1:length(vs)) .+ (last_i-1))
        last_i += length(vars)-1
      elseif v isa Symbol
        push!(allsyms, v)
      else
        error("@accepts: expected a variable name or a State tag")
      end
    end
    allsyms, i_states, prefix
  else
    error("@accepts: expected '@accepts a, b, ...'")
  end
  tuple(allsyms...), tuple(i_states...), prefix
end


function returns(ex)
  if ex isa Symbol
    (ex,)
  elseif _is_tuple(ex)
    tuple(ex.args...)
  else
    error("@returns: expected '@returns f_a, f_b, ...'")
  end
end


state(tpl, is::NTuple{N,Int64}) where N = getindex.(Ref(tpl), is)


function with_signature(ex)
  ex = MacroTools.prewalk(MacroTools.rmlines, ex)
  # display(ex)

  ex.head === :function || error("@with_signature: only works with functions")
  fn_signature = ex.args[1]
  length(fn_signature.args) == 2 || error(
    "@with_signature: function with exactly one argument required")
  fn_name, first_arg = fn_signature.args

  ex.args[2].head === :block || error("@with_singature: failed to extract function body")
  fn_body = ex.args[2].args

  # extract place holders
  length(fn_body) == 0 && error("@with_signature: missing function body")
  accepts_lines = Any[]
  for line in fn_body
    matched_accepts = MacroTools.@capture(line, @accepts accepts_variables_)
    !matched_accepts && break
    push!(accepts_lines, accepts_variables)
  end
  length(accepts_lines) == 0 && error(
    "@with_signature: expected '@accepts a, b, ...' at the start of the function body")
  matched_returns = MacroTools.@capture(fn_body[end], @returns returns_variables_)
  matched_returns || error(
    "@with_signature: expected '@returns a, b, ...' at the end of the function body")

  # @accepts and @returns can only occur right at the start or end
  if any(line -> MacroTools.@capture(line, @accepts _), fn_body[length(accepts_lines)+1:end])
    error("@with_signature: expected '@accepts a, b, ...' at the start of the function body")
  end
  for line in fn_body[1:end-1]
    MacroTools.@capture(line, @returns rest_) && error(
      "@with_signature: @returns can only occur once")
  end

  # gather variables from place holders
  accepts_vars_stateidxs_prefix = accepts.(accepts_lines)
  returns_variables = returns(returns_variables)

  # to keep backwards compatibility with State(...), but will remove State syntax the future
  for (_,stateidxs,prefix) in accepts_vars_stateidxs_prefix
    if !isnothing(prefix) && length(stateidxs) > 0
      error("cannot use State(...) and Prefix(...) together in @accepts")
    end
  end
  pos_state_variables = if length(accepts_vars_stateidxs_prefix) > 1
    for (_,stateidxs,_) in accepts_vars_stateidxs_prefix
      if length(stateidxs) > 0
        error("cannot use State(...) with @accepts arguments")
      end
    end
    ()
  else
    accepts_vars_stateidxs_prefix[1][2]
  end

  # replace the @accept lines with tuple unpacking
  allvars = Symbol[]
  for (i,(vars, stateidxs, prefix)) in enumerate(accepts_vars_stateidxs_prefix)
    if !isnothing(prefix)
      vars = Symbol.(prefix,:_,vars)
      append!(allvars, vars)
    end
    unpack = Expr(:(=), Expr(:tuple, vars...), Symbol(:args,i))
    fn_body[i] = unpack
  end

  # error on duplicated variable names (after resolving prefixes)
  if !allunique(allvars)
    uvars = unique(allvars)
    idxdups = findall(v -> count(Base.Fix1(===,v),allvars) > 1, uvars)
    dups = allvars[idxdups]
    error("duplicated arguments in all @accepts found (after resolving prefixes): $(join(dups,","))")
  end
  fn_body[end] = Expr(:tuple, returns_variables...)

  # sanitize function arguments
  (first_arg isa Expr && first_arg.head === :(::)) || error(
    "@with_signature: function argument '$first_arg' must be type annotated")
  dispatch_type = last(first_arg.args)

  # values are passed via an additional argument
  # we type annotate it here with a ::NTuple{N,Float64}
  annotated_variables_arg = Expr[]
  for (i,(vars, stateidxs, prefix)) in enumerate(accepts_vars_stateidxs_prefix)
    nvars = length(vars)
    push!(annotated_variables_arg, :($(Symbol(:args,i))::NTuple{$nvars,Float64}))
  end
  new_fn_signature = Expr(:call, fn_name, annotated_variables_arg..., first_arg)
  ex.args[1] = new_fn_signature

  # unpack the cache variable names
  accepts_variables = Tuple( vars for (vars,_,_) in accepts_vars_stateidxs_prefix )

  thismod = @__MODULE__

  new_ex = esc(quote
    $ex
    $(thismod).signature(::typeof($(fn_name)), ::$(dispatch_type)) =
        ($(accepts_variables), $(returns_variables))
    $(thismod).state_indices(::typeof($(fn_name)), ::$(dispatch_type)) = $(pos_state_variables)
    $(thismod).has_signature(::typeof($(fn_name)), ::$(dispatch_type)) = true
  end)

  # display(MacroTools.prewalk(MacroTools.rmlines, new_ex))

  return new_ex
end


#######################################################################
#                          Interface methods                          #
#######################################################################


has_signature(fn, dispatch) = false
signature(fn, dispatch) = error(
  "function '$fn' with dispatch '$(dispatch)' was not defined using @with_signature")
state_indices(fn, dispatch) = error(
  "function '$fn' with dispatch '$(dispatch)' was not defined using @with_signature")
accepts(fn, dispatch) = signature(fn, dispatch)[1]
returns(fn, dispatch) = signature(fn, dispatch)[2]


#######################################################################
#                          broadcast methods                          #
#######################################################################


"""
    broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch

Broadcast a `@with_signature` function `f` dispatched on `D` over variables of
cache `C`. The arguments and return values are determined by the `accepts, returns` overloads
of `(f, D)` from the `@with_signature` interface.

See: `@with_signature`.
"""
function broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch
  has_signature(f, D) || error("broadcast_volume!: Function '$f' has no signature!")
  accepts_syms, returns_syms = accepts(f, D)[1], returns(f, D)
  n_accepts, n_returns = length(accepts_syms), length(returns_syms)
  v = get_variable(C, first(accepts_syms))
  accepts_vars = NTuple{n_accepts,Vector{Float64}}( get_variable(C, a) for a in accepts_syms )
  returns_vars = NTuple{n_returns,Vector{Float64}}( get_variable(C, r) for r in returns_syms )
  f_args = args -> f(args, D)
  _broadcast_volume!(f_args, C, accepts_vars, returns_vars)
  return
end


function _broadcast_volume!(f, C::Cache,
    accepts_vars::NTuple{N,Vector{Float64}},
    returns_vars::NTuple{M,Vector{Float64}}) where {N,M}

  soa_accepts = StructArray{NTuple{N,Float64}}( accepts_vars )
  soa_returns = StructArray{NTuple{M,Float64}}( returns_vars )
  npts = length(first(accepts_vars))
  for idx = 1:npts
    soa_returns[idx] = f(soa_accepts[idx])
  end
  return
end


@deprecate broadcast_volume!(f, rsolver, cache, mesh) broadcast_faces!(f, rsolver, cache, mesh)


"""
    broadcast_faces!(f, rsolver::ApproxRiemannSolver, cache::Cache, mesh::Mesh)

Broadcast `@with_signature` function `f` with dispatch `rsolver` over `cache` and
faces of `mesh`.
"""
function broadcast_faces!(f, rsolver::AbstractRiemannSolver, cache::Cache, mesh::Mesh)
  has_signature(f, rsolver) || error("broadcast_faces!: Function '$f' has no signature!")
  accepts_vars = accepts(f, rsolver)
  returns_vars = returns(f, rsolver)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_face!(f_args, mesh, accepts_data, returns_data)
  return
end


function broadcast_faces!(f, equation, mesh::Mesh)
  has_signature(f, equation) || error("broadcast_faces!: Function '$f' has no signature!")
  accepts_vars = accepts(f, equation)
  returns_vars = returns(f, equation)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_face!(f_args, mesh, accepts_data, returns_data)
  return
end


function _broadcast_face!(f, mesh::Mesh1d,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}}) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx = StructArray{NTuple{Nr,Float64}}( nflx_data )
  Npts, K = layout(mesh)
  Idx = LinearIndices((Npts, K))
  n_out = (-1.0,) # 'normal' in negative direction
  for cell = 1:K
    # each loop deals with the lhs' interface of the current cell
    cell_out, cell_in = periodic_index(cell-1, K), cell
    idx_out, idx_in = Idx[end,cell_out], Idx[1,cell_in]
    accepts_out, accepts_in = soa_state[idx_out], soa_state[idx_in]
    nflx_out = f((accepts_out, accepts_in), n_out)
    # n_out * f(u_out, u_in, n_out) = - n_in * f(u_in, u_out, n_in) with n_out = -n_in
    nflx_in  = @. -nflx_out
    face_idx_rhs = cell
    face_idx_lhs = K+periodize(cell-K-1, K)
    soa_nflx[face_idx_rhs] = nflx_out
    soa_nflx[face_idx_lhs] = nflx_in
  end
  return
end


function _broadcast_face!(f, mesh::Mesh2d,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}}) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx  = StructArray{NTuple{Nr,Float64}}( nflx_data )

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  @unpack nx, ny = get_bdry_variables(mesh.cache)

  for cell = 1:ncells(mesh)

    bulk_in = cellindices(mesh, cell)
    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, cell)

    # xmin face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Xmin]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_lhs)
      idx_out = bulk_out[end,fidx]
      idx_in  = bulk_in[1,fidx]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # xmax face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Xmax]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_rhs)
      idx_out = bulk_out[1,fidx]
      idx_in  = bulk_in[end,fidx]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # ymin face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Ymin]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_down)
      idx_out = bulk_out[fidx,end]
      idx_in  = bulk_in[fidx,1]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # ymax face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Ymax]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_up)
      idx_out = bulk_out[fidx,1]
      idx_in  = bulk_in[fidx,end]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

  end
  return
end


"""
    broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh, t)

Impose boundary conditions using the flux method `f` and boundary conditions `bc`
for fields in `cache` on `mesh` at time `t`.
"""
function broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh1d, t=0.0)
  @unpack lhs_bc, rhs_bc, rsolver = bc
  has_signature(f, rsolver) || error("broadcast_volume!: Function '$f' has no signature!")
  bc.isperiodic && return

  accepts_vars = accepts(f, rsolver)
  returns_vars = returns(f, rsolver)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_boundaries!(f_args, mesh, accepts_data, returns_data, lhs_bc, rhs_bc, t)
  return
end


boundary_state(bc::AbstractBC, u_inner, t) = bc(t)
boundary_state(bc::OutflowBC, u_inner, t)  = u_inner


# similar to _broadcast_volume!, but only iterating domain bdrys and broadcasting
# over two returned data arrays
function _broadcast_boundaries!(f, mesh::Mesh,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}},
    lhs_bc::AbstractBC,
    rhs_bc::AbstractBC,
    t) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx = StructArray{NTuple{Nr,Float64}}( nflx_data )
  Npts, K = layout(mesh)

  # lhs bdry
  cell = 1
  state = soa_state[cell]
  bdry_state = boundary_state(lhs_bc, state, t)
  n_out = (-1.0,)
  nf = f((bdry_state, state), n_out)
  soa_nflx[1] = f((bdry_state, state), n_out)

  # rhs bdry
  cell = K
  state = soa_state[end]
  bdry_state = boundary_state(rhs_bc, state, t)
  n_out = (1.0,)
  soa_nflx[end] = f((bdry_state, state), n_out)

  return
end


#######################################################################
#                         new implementation                          #
#######################################################################


function wrap_variables_in_soa(cache, vars1)
  n1 = length(vars1)
  v1 = NTuple{n1,Vector{Float64}}( get_variable(cache, v) for v in vars1 )
  soa_v1 = StructArray{NTuple{n1,Float64}}( v1 )
  return soa_v1
end
function wrap_variables_in_soa(cache, vars1, vars2, vars3)
  n1 = length(vars1)
  v1 = NTuple{n1,Vector{Float64}}( get_variable(cache, v) for v in vars1 )
  soa_v1 = StructArray{NTuple{n1,Float64}}( v1 )
  n2 = length(vars2)
  v2 = NTuple{n2,Vector{Float64}}( get_variable(cache, v) for v in vars2 )
  soa_v2 = StructArray{NTuple{n2,Float64}}( v2 )
  n3 = length(vars3)
  v3 = NTuple{n3,Vector{Float64}}( get_variable(cache, v) for v in vars3 )
  soa_v3 = StructArray{NTuple{n3,Float64}}( v3 )
  return soa_v1, soa_v2, soa_v3
end


function broadcast_volume_2!(f, equation, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_volume_2!: '$f' was not defined using @with_signature"
  accepts_vars1, = accepts(f, equation)
  returns_vars = returns(f, equation)
  args = wrap_variables_in_soa(mesh.cache, accepts_vars1)
  rets = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f =  idx -> rets[idx] = f(args[idx], equation)
  _broadcast_volume_2!(wrapped_f, mesh)
  return
end


function _broadcast_volume_2!(f, m::Mesh)
  for idx = 1:length(m)
    @inline f(idx)
  end
  return
end


function broadcast_faces_2!(f, equation, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_faces!: '$f' was not defined using @with_signature"
  vars1, vars2, vars3 = accepts(f, equation)
  returns_vars = returns(f, equation)
  args1, args2, args3 = wrap_variables_in_soa(mesh.cache, vars1, vars2, vars3)
  rets  = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f = (i,j1,j2,j3) -> rets[i] = f(args1[j1],args2[j2],args3[j3],equation)
  _broadcast_face_2!(wrapped_f, mesh)
  return
end


function _broadcast_face_2!(f, mesh::Mesh2d)

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  # @unpack nx, ny = get_bdry_variables(mesh.cache)

  for (icell,cell) = enumerate(mesh.tree.cells)

    bulk_in = cellindices(mesh, icell)
    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, icell)

    # xmin face
    if has_neighbor(cell, Cart2d.Xmin)
      cell_out = mesh.tree.cells[icell].neighbors[Cart2d.Xmin]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_lhs)
        idx_out = bulk_out[end,fidx]
        idx_in  = bulk_in[1,fidx]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    # xmax face
    if has_neighbor(cell, Cart2d.Xmax)
      cell_out = mesh.tree.cells[icell].neighbors[Cart2d.Xmax]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_rhs)
        idx_out = bulk_out[1,fidx]
        idx_in  = bulk_in[end,fidx]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    # ymin face
    if has_neighbor(cell, Cart2d.Ymin)
      cell_out = mesh.tree.cells[icell].neighbors[Cart2d.Ymin]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_down)
        idx_out = bulk_out[fidx,end]
        idx_in  = bulk_in[fidx,1]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    # ymax face
    if has_neighbor(cell, Cart2d.Ymax)
      cell_out = mesh.tree.cells[icell].neighbors[Cart2d.Ymax]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_up)
        idx_out = bulk_out[fidx,1]
        idx_in  = bulk_in[fidx,end]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

  end
  return
end


function broadcast_bdry_2!(f, equation, bdrycond::AbstractBC, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_bdry!: '$f' was not defined using @with_signature"
  vars1, vars2, vars3 = accepts(f, equation)
  returns_vars = returns(f, equation)
  args1, args2, args3 = wrap_variables_in_soa(mesh.cache, vars1, vars2, vars3)
  rets  = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f = (i,j1,j2,j3) -> rets[i] = f(args1[j1],args2[j2],args3[j3],equation)
  _broadcast_bdry_2!(wrapped_f, bdrycond, mesh)
  return
end


function _broadcast_bdry_2!(f, bdrycond::DirichletBC2, mesh::Mesh2d)

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  # @unpack nx, ny = get_bdry_variables(mesh.cache)

  for bdrycell in mesh.tree.bdrycells

    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, bdrycell.index)
    idxs = cellindices(mesh, bdrycell.index)

    # @show rng_lhs, rng_rhs, rng_down, rng_up
    # TODO("need to use bulk and face indices here!")

    if !has_neighbor(bdrycell, Cart2d.Xmin)

      for (fidx, face_idx) in enumerate(rng_lhs)
        bidx = idxs[fidx,1]
        @inline f(face_idx, bidx, bidx, bidx)
      end

    elseif !has_neighbor(bdrycell, Cart2d.Xmax)

      for (fidx, face_idx) in enumerate(rng_rhs)
        bidx = idxs[fidx,end]
        @inline f(face_idx, bidx, bidx, bidx)
      end

    elseif !has_neighbor(bdrycell, Cart2d.Ymin)

      for (fidx, face_idx) in enumerate(rng_down)
        bidx = idxs[1,fidx]
        @inline f(face_idx, bidx, bidx, bidx)
      end

    elseif !has_neighbor(bdrycell, Cart2d.Ymax)

      for (fidx, face_idx) in enumerate(rng_up)
        bidx = idxs[end,fidx]
        @inline f(face_idx, bidx, bidx, bidx)
      end

    else

      error("Something went wrong ...")

    end

  end
  # TODO()
  return
end
